// Copyright (c) 2017 Electronic Theatre Controls, Inc., http://www.etcconnect.com
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.


/*******************************************************************************
 *
 *   Electronic Theatre Controls
 *
 *   lighthack - Box 2
 *
 *   (c) 2017 by ETC
 *   (c) 2018 by Stefan Staub
 *
 *
 *   This code implements a Pan/Tilt module using two encoders and three
 *   buttons. The two encoders function as pan and tilt controllers with one
 *   button being reserved for controlling fine/coarse mode. The other two
 *   buttons are assigned to Next and Last which make it easy to switch between
 *   channels.
 *
 *******************************************************************************
 *
 *  NOTE: UPDATE VERSION_STRING IN DEFINITIONS BELOW WHEN VERSION NUMBER CHANGES
 *
 *  Revision History
 *
 *  yyyy-mm-dd   Vxx      By_Who                 Comment
 *
 *  2017-07-21   1.0.0.1  Ethan Oswald Massey    Original creation
 *
 *  2017-10-19   1.0.0.2  Sam Kearney            Fix build errors on some
 *                                               Arduino platforms. Change
 *                                               OSC subscribe parameters
 *
 *  2017-10-24   1.0.0.3  Sam Kearney            Add ability to scale encoder
 *                                               output
 *
 *  2017-11-22   1.0.0.4  Hans Hinrichsen        Add splash msg before Eos
 *                                               connects
 *
 *  2017-12-07   1.0.0.5  Hans Hinrichsen        Added timeout to disconnect
 *                                               and show splash screen again
 * 
 *  2018-06-29   2.0.0.1  Stefan Staub           Modification for stepping to
 *                                               a parameter list
 *
 ******************************************************************************/

/*******************************************************************************
 * Includes
 ******************************************************************************/
#include "Arduino.h"
#include <OSCMessage.h>
#ifdef BOARD_HAS_USB_SERIAL
#include <SLIPEncodedUSBSerial.h>
SLIPEncodedUSBSerial SLIPSerial(thisBoardsSerialUSB);
#else
#include <SLIPEncodedSerial.h>
SLIPEncodedSerial SLIPSerial(Serial);
#endif
#include <LiquidCrystal.h>

/*******************************************************************************
 * Macros and Constants
 ******************************************************************************/
#define LCD_CHARS           16
#define LCD_LINES           2   // Currently assume at least 2 lines

#define LCD_RS              12
#define LCD_ENABLE          11
#define LCD_D4              5
#define LCD_D5              4
#define LCD_D6              3
#define LCD_D7              2

#define ENC_1_A             A0
#define ENC_1_B             A1
#define ENC_2_A             A2
#define ENC_2_B             A3
#define UP_BTN              6
#define DOWN_BTN            7
#define SHIFT_BTN           8

#define SUBSCRIBE           1
#define UNSUBSCRIBE         0

#define UP                  0
#define DOWN                1

// number of encoders you use
#define ENCODERS            2

#define FORWARD             0
#define REVERSE             1

// Change these values to switch which direction increase/decrease pan/tilt
#define ENC_1_DIR   FORWARD
#define ENC_2_DIR   FORWARD

// Use these values to make the encoder more coarse or fine. This controls
// the number of wheel "ticks" the device sends to Eos for each tick of the
// encoder. 1 is the default and the most fine setting. Must be an integer.
#define WHEEL_ACC           4   // Used for intens parameter instead of fine

#define SIG_DIGITS          3   // Number of significant digits displayed

#define OSC_BUF_MAX_SIZE    512

const String VERSION_STRING = "2.0.0.1";
const String HANDSHAKE_QUERY = "ETCOSC?";
const String HANDSHAKE_REPLY = "OK";
const String PING_QUERY = "box2_hello";
const String SUBSCRIBE_QUERY = "/eos/subscribe/param/";
const String UNSUBSCRIBE_QUERY = "/eos/subscribe/param/*";
const String PARAMETER_QUERY = "eos/out/param/";

// Change these values to alter how long we wait before sending an OSC ping
// to see if Eos is still there, and then finally how long before we
// disconnect and show the splash screen
// Values are in milliseconds
#define PING_AFTER_IDLE_INTERVAL    2500
#define TIMEOUT_AFTER_IDLE_INTERVAL 5000

/*******************************************************************************
 * Local Types
 ******************************************************************************/
struct Parameter {
    String name;
    String display;
  float value;
};

const uint8_t PARAMETER_MAX = 16; // number of parameters must even
struct Parameter parameter[PARAMETER_MAX] = 
{
    {"empty", "------"},
    {"Intens"},
    {"Pan"},
    {"Tilt"},
    {"Zoom"},
    {"Edge"},
    {"Iris"},
    {"Diffusn"},
    {"Hue"},
    {"Saturatn"},
    {"Red"},
    {"Green"},
    {"Blue"},
    {"Cyan"},
    {"Magenta"},
    {"Yellow"}/*,
    {"cto", "CTO"},
    {"frame_assembly", "Assembly"},
    {"thrust_A", "ThrustA"},
    {"angle_A", "AngleA"},
    {"thrust_B", "ThrustB"},
    {"angle_B", "AngleB"},
    {"thrust_C", "ThrustC"},
    {"angle_C", "AngleC"},
    {"thrust_D", "ThrustD"},
    {"angle_D", "AngleD"}*/ 
};

struct Encoder
{
    uint8_t parameterIdx;
    uint8_t pinA;
    uint8_t pinB;
    uint8_t pinAPrevious;
    uint8_t pinBPrevious;
    uint8_t direction;
};
struct Encoder Encoder1;
struct Encoder Encoder2;

struct Button
{
    uint8_t function;
    uint8_t pin;
    uint8_t last;
}; 
struct Button parameterUp;
struct Button parameterDown;


/*******************************************************************************
 * Global Variables
 ******************************************************************************/

// initialize the library with the numbers of the interface pins
LiquidCrystal lcd(LCD_RS, LCD_ENABLE, LCD_D4, LCD_D5, LCD_D6, LCD_D7);

bool updateDisplay = false;
bool connectedToEos = false;
unsigned long lastMessageRxTime = 0;
bool timeoutPingSent = false;
int8_t index = 2; // start with parameter index 2 must even

/*******************************************************************************
 * Local Functions
 ******************************************************************************/

/*******************************************************************************
 * Send a ping to eos to check if you are alive
 *
 * Parameters:  none
 *
 * Return Value: void
 *
 ******************************************************************************/
void sendPing()
{
    OSCMessage ping("/eos/ping");
    ping.add(PING_QUERY.c_str());
    SLIPSerial.beginPacket();
    ping.send(SLIPSerial);
    SLIPSerial.endPacket();
    timeoutPingSent = true;
}

/*******************************************************************************
 * Issues all our filters to Eos. 
 *
 * Parameters:  none
 *
 * Return Value: void
 *
 ******************************************************************************/
void issueFilters()
{
    // Add a filter so we don't get spammed with unwanted OSC messages from Eos
    OSCMessage filter("/eos/filter/add");
    filter.add("/eos/out/param/*");
    filter.add("/eos/out/ping");
    SLIPSerial.beginPacket();
    filter.send(SLIPSerial);
    SLIPSerial.endPacket();
}

/*******************************************************************************
 * Issues all our subscribes to Eos. When subscribed, Eos will keep us updated
 * with the latest values for a given parameter.
 *
 * Parameters:  none
 *
 * Return Value: void
 *
 ******************************************************************************/
void issueSubscribes()
{
    String unsubMsg = UNSUBSCRIBE_QUERY;
    OSCMessage unsubscribe(unsubMsg.c_str());
    unsubscribe.add(UNSUBSCRIBE);
    SLIPSerial.beginPacket();
    unsubscribe.send(SLIPSerial);
    SLIPSerial.endPacket();

    // subscribes all parameters of your list, exept item 0
    String subMsg = SUBSCRIBE_QUERY + parameter[index].name + "/" + parameter[index + 1].name;
    OSCMessage subscribe(subMsg.c_str());
    subscribe.add(SUBSCRIBE);
    SLIPSerial.beginPacket();
    subscribe.send(SLIPSerial);
    SLIPSerial.endPacket();
}

/*******************************************************************************
 * Init the console, gives back a handshake reply and send the subscribtions
 *
 * Parameters:  none
 *
 * Return Value: void
 *
 ******************************************************************************/
void initEOS()
{
    SLIPSerial.beginPacket();
    SLIPSerial.write((const uint8_t*)HANDSHAKE_REPLY.c_str(), (size_t)HANDSHAKE_REPLY.length());
    SLIPSerial.endPacket();
    issueFilters();
    issueSubscribes();
}

/*******************************************************************************
 * Given an unknown OSC message we check to see if it's a handshake message.
 * If it's a handshake we issue a subscribe, otherwise we begin to check
 * if there is some data for the parameters of our list.
 *
 * Parameters: msg - The OSC message of unknown importance
 *
 * Return Value: void
 *
 ******************************************************************************/
void parseOSCMessage(String& msg)
{
    // check to see if this is the handshake string
    if (msg.indexOf(HANDSHAKE_QUERY) != -1)
    {
        initEOS();
        // Make our splash screen go away
        connectedToEos = true;
        updateDisplay = true;
    }
    else
    {
        // checks if there is a message with data of your parameter list
        OSCMessage oscmsg;
        oscmsg.fill((uint8_t*)msg.c_str(), (int)msg.length());

        String parseMsgA = PARAMETER_QUERY + parameter[index].name;
            if(msg.indexOf(parseMsgA) != -1)
            {
                parameter[index].value = oscmsg.getOSCData(0)->getFloat(); // get the value
                connectedToEos = true; // Update this here just in case we missed the handshake
                updateDisplay = true;
                return;
            }
        
        String parseMsgB = PARAMETER_QUERY + parameter[index + 1].name;
            if(msg.indexOf(parseMsgB) != -1)
            {
                parameter[index + 1].value = oscmsg.getOSCData(0)->getFloat(); // get the value
                connectedToEos = true; // Update this here just in case we missed the handshake
                updateDisplay = true;
            }
    }
}

/*******************************************************************************
 * Updates the display with the latest pan and tilt positions.
 *
 * Parameters:  none
 *
 * Return Value: void
 *
 ******************************************************************************/
void displayStatus()
{
    lcd.clear();

    if (!connectedToEos)
    {
      // display a splash message before the Eos connection is open
      lcd.setCursor(0, 0);
      lcd.print("Box2 v" + VERSION_STRING);
      lcd.setCursor(0, 1);
      lcd.print("waiting for eos");
    }
    else
    {
        // put the cursor at the begining of the first line
        lcd.setCursor(0, 0);
        // check if display name is empty, if thrue take the parameter name
        if (parameter[Encoder1.parameterIdx].display.length() == 0)
        {
            lcd.print(parameter[Encoder1.parameterIdx].name);
        }
        else
        {
            lcd.print(parameter[Encoder1.parameterIdx].display);
        }
        lcd.setCursor(8, 0);
        if (parameter[Encoder2.parameterIdx].display.length() == 0)
        {
            lcd.print(parameter[Encoder2.parameterIdx].name);
        }
        else 
        {
            lcd.print(parameter[Encoder2.parameterIdx].display);
        }

    // put the cursor at the begining of the second line
    lcd.setCursor(0, 1);
    lcd.print(parameter[Encoder1.parameterIdx].value, SIG_DIGITS);
    lcd.setCursor(8, 1);
    lcd.print(parameter[Encoder2.parameterIdx].value, SIG_DIGITS);
    }
    updateDisplay = false;
}

/*******************************************************************************
 * Initializes a given encoder struct to the requested parameters.
 *
 * Parameters:
 *  encoder - Pointer to the encoder we will be initializing
 *  pinA - Where the A pin is connected to the Arduino
 *  pinB - Where the B pin is connected to the Arduino
 *  direction - Determines if clockwise or counterclockwise is "forward"
 *
 * Return Value: void
 *
 ******************************************************************************/
void initEncoder(struct Encoder* encoder, uint8_t pinA, uint8_t pinB, uint8_t direction)
{
    encoder->pinA = pinA;
    encoder->pinB = pinB;
    encoder->direction = direction;

    pinMode(pinA, INPUT_PULLUP);
    pinMode(pinB, INPUT_PULLUP);

    encoder->pinAPrevious = digitalRead(pinA);
    encoder->pinBPrevious = digitalRead(pinB);
}

/*******************************************************************************
 * Checks if the encoder has moved by comparing the previous state of the pins
 * with the current state. If they are different, we know there is movement.
 * In the event of movement we update the current state of our pins.
 *
 * Parameters:
 *  encoder - Pointer to the encoder we will be checking for motion
 *
 * Return Value: void
 *
 ******************************************************************************/
void updateEncoder(struct Encoder* encoder)
{
    int8_t encoderMotion = 0;
    int pinACurrent = digitalRead(encoder->pinA);
    int pinBCurrent = digitalRead(encoder->pinB);

    // has the encoder moved at all?
    if (encoder->pinAPrevious != pinACurrent)
    {
        // Since it has moved, we must determine if the encoder has moved forwards or backwards
        encoderMotion = (encoder->pinAPrevious == encoder->pinBPrevious) ? -1 : 1;

        // If we are in reverse mode, flip the direction of the encoder motion
        if (encoder->direction == REVERSE)
            encoderMotion = -encoderMotion;
    }
    encoder->pinAPrevious = pinACurrent;
    encoder->pinBPrevious = pinBCurrent;
    // send the updated values of the given parameter 
    if ((encoderMotion != 0) && (encoder->parameterIdx > 0))
    {
        String wheelMsg("/eos/wheel");
        if (encoder->parameterIdx == 1) // 1 = intensity
        {
            // use accelarion for the intensity wheel instead of fine
            if (digitalRead(SHIFT_BTN) == LOW) encoderMotion *= WHEEL_ACC;
        }
        else if (encoder->parameterIdx > 1)
        {
            if (digitalRead(SHIFT_BTN) == LOW) wheelMsg.concat("/fine");
            else wheelMsg.concat("/coarse");
        }
    wheelMsg.concat('/' + parameter[encoder->parameterIdx].name);
    OSCMessage wheelUpdate(wheelMsg.c_str());
    wheelUpdate.add(encoderMotion);
    SLIPSerial.beginPacket();
    wheelUpdate.send(SLIPSerial);
    SLIPSerial.endPacket();
    }
}

/*******************************************************************************
 * Initializes a given button struct to the requested parameters.
 *
 * Parameters:
 *  button - Pointer to the encoder we will be initializing
 *  pin - Where the pin is connected to the Arduino
 *  function - button function UP or DOWN
 *
 * Return Value: void
 *
 ******************************************************************************/
void initButton(struct Button* button, uint8_t pin, uint8_t function)
{
    button->function = function;
    button->pin = pin;
    button->last = HIGH;
    pinMode(pin, INPUT_PULLUP);
}

void updateButton(struct Button* button)
{
    if ((digitalRead(button->pin)) != button->last) 
    {
        if (button->last == LOW) 
        {
            button->last = HIGH;
            //calc index
            if (button->function == UP) 
            {
                index += ENCODERS;
                if (index > PARAMETER_MAX - ENCODERS) 
                {
                    index = 0;
                } 
                Encoder1.parameterIdx = index;
                Encoder2.parameterIdx = index + 1;
                issueSubscribes();
                displayStatus();
            }

        if(button->function == DOWN) 
        {
            index -= ENCODERS;
            if(index < 0) 
            {
                index = PARAMETER_MAX - ENCODERS;
            } 
            Encoder1.parameterIdx = index;
            Encoder2.parameterIdx = index + 1;
            issueSubscribes();
            displayStatus();
            }      
        }
        else 
        {
            button->last = LOW;
        }
    }
}

/*******************************************************************************
 * Here we setup our encoder, lcd, and various input devices. We also prepare
 * to communicate OSC with Eos by setting up SLIPSerial. Once we are done with
 * setup() we pass control over to loop() and never call setup() again.
 *
 * NOTE: This function is the entry function. This is where control over the
 * Arduino is passed to us (the end user).
 *
 * Parameters: none
 *
 * Return Value: void
 *
 ******************************************************************************/
void setup()
{
    SLIPSerial.begin(115200);
    // This is a hack around an Arduino bug. It was taken from the OSC library
    //examples
#ifdef BOARD_HAS_USB_SERIAL
    while (!SerialUSB);
#else
    while (!Serial);
#endif

    lcd.begin(LCD_CHARS, LCD_LINES);
    lcd.clear();

    initEOS(); // for hotplug with Arduinos without native USB like UNO

    // init of hardware elements
    initEncoder(&Encoder1, A0, A1, ENC_1_DIR);
    initEncoder(&Encoder2, A2, A3, ENC_2_DIR);
    Encoder1.parameterIdx = index;
    Encoder2.parameterIdx = index + 1;

    pinMode(SHIFT_BTN, INPUT_PULLUP);

    initButton(&parameterUp, UP_BTN, UP);
    initButton(&parameterDown, DOWN_BTN, DOWN);

    displayStatus();
}

/*******************************************************************************
 * Here we service, monitor, and otherwise control all our peripheral devices.
 * First, we retrieve the status of our encoders and buttons and update Eos.
 * Next, we check if there are any OSC messages for us.
 * Finally, we update our display (if an update is necessary)
 *
 * NOTE: This function is our main loop and thus this function will be called
 * repeatedly forever
 *
 * Parameters: none
 *
 * Return Value: void
 *
 ******************************************************************************/
void loop()
{
    static String curMsg;
    int size;

    // Then we check to see if any OSC commands have come from Eos
    // and update the display accordingly.
    size = SLIPSerial.available();
    if (size > 0)
    {
        // Fill the msg with all of the available bytes
        while (size--)
            curMsg += (char)(SLIPSerial.read());
    }
    if (SLIPSerial.endofPacket())
    {
        parseOSCMessage(curMsg);
        lastMessageRxTime = millis();
        // We only care about the ping if we haven't heard recently
        // Clear flag when we get any traffic
        timeoutPingSent = false;
        curMsg = String();
    }

    // check for encoder and up/down updates
    updateEncoder(&Encoder1);
    updateEncoder(&Encoder2);
    updateButton(&parameterUp);
    updateButton(&parameterDown);

    if(lastMessageRxTime > 0) 
    {
        unsigned long diff = millis() - lastMessageRxTime;
        //We first check if it's been too long and we need to time out
        if(diff > TIMEOUT_AFTER_IDLE_INTERVAL) 
        {
            connectedToEos = false;
            lastMessageRxTime = 0;
            updateDisplay = true;
            timeoutPingSent = false;
        }

        //It could be the console is sitting idle. Send a ping once to
        // double check that it's still there, but only once after 2.5s have passed
        if(!timeoutPingSent && diff > PING_AFTER_IDLE_INTERVAL) 
        {
            sendPing();
        }
    }

    if (updateDisplay)
        displayStatus();
}
